const Stmt = @import("./Stmt.zig");

const coral = @import("coral");

const tokens = @import("./tokens.zig");

const tree = @import("./tree.zig");

next: ?*const Self = null,
line: u32,

kind: union (enum) {
	nil_literal,
	true_literal,
	false_literal,
	number_literal: []const coral.io.Byte,
	string_literal: []const coral.io.Byte,
	symbol_literal: []const coral.io.Byte,
	table_construct: TableConstruct,
	key_value: KeyValue,
	group: *Self,
	lambda_construct: LambdaConstruct,
	declaration_get: DeclarationGet,
	declaration_set: DeclarationSet,
	field_get: FieldGet,
	field_set: FieldSet,
	subscript_get: SubscriptGet,
	subscript_set: SubscriptSet,
	binary_op: BinaryOp,
	unary_op: UnaryOp,
	invoke: Invoke,
	import_builtin,
	print_builtin,
	vec2_builtin,
	vec3_builtin,
},

pub const BinaryOp = struct {
	rhs_operand: *Self,
	lhs_operand: *Self,
	operation: Operation,

	pub const Operation = enum {
		addition,
		subtraction,
		multiplication,
		divsion,
		equals_comparison,
		greater_than_comparison,
		greater_equals_comparison,
		less_than_comparison,
		less_equals_comparison,
	};

	fn parser(comptime parse_next: Parser, comptime operations: []const BinaryOp.Operation) Parser {
		const BinaryOpParser = struct {
			fn parse(root: *tree.Root, stream: *tokens.Stream, environment: *tree.Environment) tree.ParseError!*Self {
				var expression = try parse_next(root, stream, environment);

				inline for (operations) |operation| {
					const token = comptime @as(tokens.Token, switch (operation) {
						.addition => .symbol_plus,
						.subtraction => .symbol_minus,
						.multiplication => .symbol_asterisk,
						.divsion => .symbol_forward_slash,
						.equals_comparison => .symbol_double_equals,
						.greater_than_comparison => .symbol_greater_than,
						.greater_equals_comparison => .symbol_greater_equals,
						.less_than_comparison => .symbol_less_than,
						.less_equals_comparison => .symbol_less_equals,
					});

					if (stream.token == coral.io.tag_of(token)) {
						stream.step();

						if (stream.token == .end) {
							return root.report_error(stream, "expected other half of expression after `" ++ comptime token.text() ++ "`", .{});
						}

						// TODO: Remove once Zig has fixed struct self-reassignment.
						const unnecessary_temp = expression;

						expression = try root.create_expr(.{
							.line = stream.lines_stepped,

							.kind = .{
								.binary_op = .{
									.rhs_operand = try parse_next(root, stream, environment),
									.operation = operation,
									.lhs_operand = unnecessary_temp,
								},
							},
						});
					}
				}

				return expression;
			}
		};

		return BinaryOpParser.parse;
	}
};

pub const DeclarationGet = struct {
	declaration: *const tree.Declaration,
};

pub const DeclarationSet = struct {
	declaration: *const tree.Declaration,
	assign: *const Self,
};

pub const FieldGet = struct {
	identifier: []const coral.io.Byte,
	object: *const Self,
};

pub const FieldSet = struct {
	identifier: []const coral.io.Byte,
	object: *const Self,
	assign: *const Self,
};

pub const Invoke = struct {
	argument: ?*const Self,
	object: *const Self,
};

pub const KeyValue = struct {
	key: *const Self,
	value: *const Self,
};

pub const LambdaConstruct = struct {
	environment: *const tree.Environment,
};

const Parser = fn (root: *tree.Root, stream: *tokens.Stream, environment: *tree.Environment) tree.ParseError!*Self;

const Self = @This();

pub const SubscriptGet = struct {
	index: *const Self,
	object: *const Self,
};

pub const SubscriptSet = struct {
	index: *const Self,
	object: *const Self,
	assign: *const Self,
};

pub const TableConstruct = struct {
	entry: ?*const Self,
};

pub const UnaryOp = struct {
	operand: *Self,
	operation: Operation,

	pub const Operation = enum {
		numeric_negation,
		boolean_negation,
	};
};

pub fn parse(root: *tree.Root, stream: *tokens.Stream, environment: *tree.Environment) tree.ParseError!*Self {
	const expression = try parse_additive(root, stream, environment);

	if (stream.token == .symbol_equals) {
		stream.skip_newlines();

		if (stream.token == .end) {
			return root.report_error(stream, "expected assignment after `=`", .{});
		}

		return root.create_expr(.{
			.line = stream.lines_stepped,

			.kind = switch (expression.kind) {
				.declaration_get => |declaration_get| convert: {
					if (declaration_get.declaration.is.readonly) {
						return root.report_error(stream, "readonly declarations cannot be re-assigned", .{});
					}

					break: convert .{
						.declaration_set = .{
							.assign = try parse(root, stream, environment),
							.declaration = declaration_get.declaration,
						},
					};
				},

				.field_get => |field_get| .{
					.field_set = .{
						.assign = try parse(root, stream, environment),
						.object = field_get.object,
						.identifier = field_get.identifier,
					},
				},

				.subscript_get => |subscript_get| .{
					.subscript_set = .{
						.assign = try parse(root, stream, environment),
						.object = subscript_get.object,
						.index = subscript_get.index,
					},
				},

				else => return root.report_error(stream, "expected local or field on left-hand side of expression", .{}),
			},
		});
	}

	return expression;
}

const parse_additive = BinaryOp.parser(parse_equality, &.{
	.addition,
	.subtraction,
});

const parse_comparison = BinaryOp.parser(parse_term, &.{
	.greater_than_comparison,
	.greater_equals_comparison,
	.less_than_comparison,
	.less_equals_comparison
});

const parse_equality = BinaryOp.parser(parse_comparison, &.{
	.equals_comparison,
});

fn parse_factor(root: *tree.Root, stream: *tokens.Stream, environment: *tree.Environment) tree.ParseError!*Self {
	var expression = try parse_operand(root, stream, environment);

	while (true) {
		switch (stream.token) {
			.symbol_period => {
				stream.skip_newlines();

				// TODO: Remove when Zig fixes miscompilation with in-place struct re-assignment.
				const unnecessary_temp = expression;

				expression = try root.create_expr(.{
					.line = stream.lines_stepped,

					.kind = .{
						.field_get = .{
							.identifier = switch (stream.token) {
								.identifier => |field_identifier| field_identifier,
								else => return root.report_error(stream, "expected identifier after `.`", .{}),
							},

							.object = unnecessary_temp,
						},
					},
				});

				stream.skip_newlines();
			},

			.symbol_bracket_left => {
				stream.skip_newlines();

				// TODO: Remove when Zig fixes miscompilation with in-place struct re-assignment.
				const unnecessary_temp = expression;

				expression = try root.create_expr(.{
					.line = stream.lines_stepped,

					.kind = .{
						.subscript_get = .{
							.index = try parse(root, stream, environment),
							.object = unnecessary_temp,
						},
					},
				});

				if (stream.token != .symbol_bracket_right) {
					return root.report_error(stream, "expected closing `]` on subscript", .{});
				}

				stream.skip_newlines();
			},

			.symbol_paren_left => {
				const lines_stepped = stream.lines_stepped;

				stream.skip_newlines();

				var first_argument = @as(?*Self, null);

				if (stream.token != .symbol_paren_right) {
					var argument = try parse(root, stream, environment);

					first_argument = argument;

					while (true) {
						switch (stream.token) {
							.symbol_comma => stream.skip_newlines(),
							.symbol_paren_right => break,
							else => return root.report_error(stream, "expected `,` or `)` after lambda argument", .{}),
						}

						const next_argument = try parse(root, stream, environment);

						argument.next = next_argument;
						argument = next_argument;
					}
				}

				stream.skip_newlines();

				// TODO: Remove when Zig fixes miscompilation with in-place struct re-assignment.
				const unnecessary_temp = expression;

				expression = try root.create_expr(.{
					.line = lines_stepped,

					.kind = .{
						.invoke = .{
							.argument = first_argument,
							.object = unnecessary_temp,
						},
					},
				});
			},

			else => break,
		}
	}

	return expression;
}

fn parse_operand(root: *tree.Root, stream: *tokens.Stream, environment: *tree.Environment) tree.ParseError!*Self {
	switch (stream.token) {
		.symbol_paren_left => {
			stream.skip_newlines();

			const expression = try parse(root, stream, environment);

			if (stream.token != .symbol_paren_right) {
				return root.report_error(stream, "expected a closing `)` after expression", .{});
			}

			stream.skip_newlines();

			return root.create_expr(.{
				.line = stream.lines_stepped,
				.kind = .{.group = expression},
			});
		},

		.keyword_nil => {
			stream.skip_newlines();

			return root.create_expr(.{
				.line = stream.lines_stepped,
				.kind = .nil_literal,
			});
		},

		.keyword_true => {
			stream.skip_newlines();

			return root.create_expr(.{
				.line = stream.lines_stepped,
				.kind = .true_literal,
			});
		},

		.keyword_false => {
			stream.skip_newlines();

			return root.create_expr(.{
				.line = stream.lines_stepped,
				.kind = .false_literal,
			});
		},

		.number => |value| {
			stream.skip_newlines();

			return root.create_expr(.{
				.line = stream.lines_stepped,
				.kind = .{.number_literal = value},
			});
		},

		.string => |value| {
			stream.skip_newlines();

			return root.create_expr(.{
				.line = stream.lines_stepped,
				.kind = .{.string_literal = value},
			});
		},

		.symbol_at => {
			stream.step();

			const identifier = switch (stream.token) {
				.identifier => |identifier| identifier,
				else => return root.report_error(stream, "expected identifier after `@`", .{}),
			};

			stream.skip_newlines();

			if (coral.io.are_equal(identifier, "import")) {
				return root.create_expr(.{
					.line = stream.lines_stepped,
					.kind = .import_builtin,
				});
			}

			if (coral.io.are_equal(identifier, "print")) {
				return root.create_expr(.{
					.line = stream.lines_stepped,
					.kind = .print_builtin,
				});
			}

			if (coral.io.are_equal(identifier, "vec2")) {
				return root.create_expr(.{
					.line = stream.lines_stepped,
					.kind = .vec2_builtin,
				});
			}

			if (coral.io.are_equal(identifier, "vec3")) {
				return root.create_expr(.{
					.line = stream.lines_stepped,
					.kind = .vec3_builtin,
				});
			}

			return root.report_error(stream, "unexpected identifier after `@`", .{});
		},

		.symbol_period => {
			stream.step();

			const identifier = switch (stream.token) {
				.identifier => |identifier| identifier,
				else => return root.report_error(stream, "expected identifier after `.`", .{}),
			};

			stream.skip_newlines();

			return root.create_expr(.{
				.line = stream.lines_stepped,
				.kind = .{.symbol_literal = identifier},
			});
		},

		.identifier => |identifier| {
			stream.skip_newlines();

			return root.create_expr(.{
				.line = stream.lines_stepped,

				.kind = .{
					.declaration_get = .{
						.declaration = (try environment.resolve_declaration(identifier)) orelse {
							return root.report_error(stream, "undefined identifier `{identifier}`", .{
								.identifier = identifier,
							});
						}
					},
				},
			});
		},

		.keyword_lambda => {
			stream.skip_newlines();

			if (stream.token != .symbol_paren_left) {
				return root.report_error(stream, "expected `(` after opening lambda block", .{});
			}

			stream.skip_newlines();

			var lambda_environment = try environment.create_enclosed(root);

			while (stream.token != .symbol_paren_right) {
				const identifier = switch (stream.token) {
					.identifier => |identifier| identifier,
					else => return root.report_error(stream, "expected identifier", .{}),
				};

				_ = lambda_environment.declare_argument(identifier) catch |declare_error| {
					return root.report_declare_error(stream, identifier, declare_error);
				};

				stream.skip_newlines();

				switch (stream.token) {
					.symbol_comma => stream.skip_newlines(),
					.symbol_paren_right => break,
					else => return root.report_error(stream, "expected `,` or `)` after identifier", .{}),
				}
			}

			stream.skip_newlines();

			if (stream.token != .symbol_colon) {
				return root.report_error(stream, "expected `:` after closing `)` of lambda identifiers", .{});
			}

			stream.skip_newlines();

			if (stream.token != .keyword_end) {
				const first_statement = try Stmt.parse(root, stream, lambda_environment);
				var current_statement = first_statement;

				while (stream.token != .keyword_end) {
					const next_statement = try Stmt.parse(root, stream, lambda_environment);

					current_statement.next = next_statement;
					current_statement = next_statement;
				}

				lambda_environment.statement = first_statement;
			}

			stream.skip_newlines();

			return root.create_expr(.{
				.line = stream.lines_stepped,
				.kind = .{.lambda_construct = .{.environment = lambda_environment}},
			});
		},

		.symbol_brace_left => {
			stream.skip_newlines();

			if (stream.token == .symbol_brace_right) {
				return root.create_expr(.{
					.line = stream.lines_stepped,
					.kind = .{.table_construct = .{.entry = null}},
				});
			}

			const first_entry = try parse_table_entry(root, stream, environment);
			var entry = first_entry;

			while (stream.token == .symbol_comma) {
				stream.skip_newlines();

				if (stream.token == .symbol_brace_right) {
					break;
				}

				const expression = try parse_table_entry(root, stream, environment);

				entry.next = expression;
				entry = expression;
			}

			if (stream.token != .symbol_brace_right) {
				return root.report_error(stream, "expected closing `}` on table construct", .{});
			}

			stream.skip_newlines();

			return root.create_expr(.{
				.line = stream.lines_stepped,
				.kind = .{.table_construct = .{.entry = first_entry}},
			});
		},

		.symbol_minus => {
			stream.skip_newlines();

			return root.create_expr(.{
				.line = stream.lines_stepped,

				.kind = .{
					.unary_op = .{
						.operand = try parse_factor(root, stream, environment),
						.operation = .numeric_negation,
					},
				},
			});
		},

		.symbol_bang => {
			stream.skip_newlines();

			return root.create_expr(.{
				.line = stream.lines_stepped,

				.kind = .{
					.unary_op = .{
						.operand = try parse_factor(root, stream, environment),
						.operation = .boolean_negation,
					},
				},
			});
		},

		else => return root.report_error(stream, "unexpected token in expression", .{}),
	}
}

fn parse_table_entry(root: *tree.Root, stream: *tokens.Stream, environment: *tree.Environment) tree.ParseError!*Self {
	switch (stream.token) {
		.symbol_period => {
			stream.step();

			const field = switch (stream.token) {
				.identifier => |identifier| identifier,
				else => return root.report_error(stream, "expected identifier in field symbol literal", .{}),
			};

			stream.skip_newlines();

			if (stream.token != .symbol_equals) {
				return root.report_error(stream, "expected `=` after table symbol key", .{});
			}

			stream.skip_newlines();

			return root.create_expr(.{
				.line = stream.lines_stepped,

				.kind = .{
					.key_value = .{
						.value = try parse(root, stream, environment),

						.key = try root.create_expr(.{
							.line = stream.lines_stepped,
							.kind = .{.symbol_literal = field},
						}),
					},
				},
			});
		},

		.symbol_bracket_left => {
			stream.skip_newlines();

			const key = try parse(root, stream, environment);

			if (stream.token != .symbol_bracket_right)  {
				return root.report_error(stream, "expected `]` after subscript index expression", .{});
			}

			stream.skip_newlines();

			if (stream.token != .symbol_equals) {
				return root.report_error(stream, "expected `=` after table expression key", .{});
			}

			stream.skip_newlines();

			return root.create_expr(.{
				.line = stream.lines_stepped,

				.kind = .{
					.key_value = .{
						.value = try parse(root, stream, environment),
						.key = key,
					},
				},
			});
		},

		else => return parse(root, stream, environment),
	}
}

const parse_term = BinaryOp.parser(parse_factor, &.{
	.multiplication,
	.divsion,
});
